"""
Data and formalisms from FAO56 related to crop evapotranspiration
"""
# {# pkglts, src
# FYEO
# #}
# {# pkglts, version, after src
from . import version

__version__ = version.__version__
# #}

from .clean import eto_daily, eto_hourly, rs_daily, rs_hourly, sat_vap, vpd



from math import acos, cos, exp, pi, sin, sqrt, tan

gsc = 0.082  # [MJ.m-2.min-1] solar constant


def eq7(z):
    """Atmospheric pressure.

    Args:
        z (float): [m] altitude above sea level

    Returns:
        P (float): [kPa] atmospheric pressure
    """
    return 101.3 * ((293 - 0.0065 * z) / 293) ** 5.26


def eq8(p):
    """Psychrometric constant

    Args:
        p (float): [kPa] atmospheric pressure

    Returns:
        gamma (float): [kPa.°C-1]
    """
    return 0.665e-3 * p


def eq11(temp):
    """Mean saturation vapour pressure

    Args:
        temp (float): [°C] air temperature

    Returns:
        es (float): [kPa]
    """
    return 0.6108 * exp(17.27 * temp / (temp + 237.3))


def eq13(temp):
    """Slope of saturation vapour pressure curve

    Args:
        temp (float): [°C] air temperature

    Returns:
        delta (float): [kPa.°C-1]
    """
    return 4098 * eq11(temp) / (temp + 237.3) ** 2


def eq21(dr, sunset_hour, latitude, delta):
    """Extraterrestrial radiation for daily periods.

    Args:
        dr (float): [-] inverse relative distance earth sun
        sunset_hour (float): [rad] sunset hour angle
        latitude (float): [rad]
        delta (float): [rad] solar declination

    Returns:
        Ra (float): [MJ.m-2.day-1]
    """
    corr = sunset_hour * sin(latitude) * sin(delta) + cos(latitude) * cos(delta) * sin(sunset_hour)
    return 24 * 60 / pi * gsc * dr * corr


def eq23(j):
    """Inverse relative distance Earth-Sun.

    Args:
        j (float): [-] number of the day in the year ranging from 1 (1 January)
                   to 365 or 366 (31 December)

    Returns:
        dr (float)
    """
    return 1 + 0.033 * cos(2 * pi * j / 365)


def eq24(j):
    """Solar declination.

    Args:
        j (float): [-] number of the day in the year ranging from 1 (1 January)
                   to 365 or 366 (31 December)

    Returns:
        delta (float)
    """
    return 0.409 * sin(2 * pi * j / 365 - 1.39)


def eq25(latitude, delta):
    """Sunset hour angle

    Args:
        latitude (float): [rad]
        delta (float): [-] solar declination

    Returns:
        omega_s (float): [rad]
    """
    return acos(-tan(latitude) * tan(delta))


def eq28(dr, omega1, omega2, latitude, delta):
    """Extraterrestrial radiation for daily periods.

    Args:
        dr (float): [-] inverse relative distance earth sun
        omega1 (float): [rad] solar time angle at beginning of period
        omega2 (float): [rad] solar time angle at end of period
        latitude (float): [rad]
        delta (float): [rad] solar declination

    Returns:
        Ra (float): [MJ.m-2.h-1]
    """
    corr = (omega2 - omega1) * sin(latitude) * sin(delta) + cos(latitude) * cos(delta) * (sin(omega2) - sin(omega1))
    return 12 * 60 / pi * gsc * dr * corr


def eq29(omega, t_length):
    """Solar time angle at beginning of period.

    Args:
        omega (float): [rad] solar time angle at midpoint of period
        t_length (float): [h] length of period

    Returns:
        omega1 (float): [rad]
    """
    return omega - pi * t_length / 24


def eq30(omega, t_length):
    """Solar time angle at end of period.

    Args:
        omega (float): [rad] solar time angle at midpoint of period
        t_length (float): [h] length of period

    Returns:
        omega1 (float): [rad]
    """
    return omega + pi * t_length / 24


def eq31(t, longitude_zone, longitude, sc):
    """Solar time angle at given time.

    Args:
        t (float): [h] standard clock time
        longitude_zone (float): [°] Longitude of the centre of the local time zone
        longitude (float): [°] longitude of measurement site
        sc (float): [h] seasonal correction for solar time

    Returns:
        omega (float): [rad]
    """
    return pi / 12 * ((t + 0.06667 * (longitude_zone - longitude) + sc) - 12)


def eq32(j):
    """Seasonal correction for solar time.

    Args:
        j (float): [-] number of the day in the year ranging from 1 (1 January)
                   to 365 or 366 (31 December)

    Returns:
        sc (float): [h]
    """
    b = 2 * pi * (j - 81) / 364  # eq33
    return 0.1645 * sin(2 * b) - 0.1255 * cos(b) - 0.025 * sin(b)


def eq34(omega_s):
    """Daylight hours.

    Args:
        omega_s (float): [rad] sunset hour angle

    Returns:
        N (float): [h]
    """
    return 24 * omega_s / pi


def eq37(ra, z):
    """Clear sky solar radiation.

    Args:
        ra (float): [MJ.m-2.day-1] Extraterrestrial radiation for daily periods
        z (float): [m] station elevation above sea level

    Returns:
        rso (float)
    """
    return (0.75 + 2e-5 * z) * ra


def eq38(rs, albedo=0.23):
    """Net shortwave radiation.

    Args:
        rs (float): [MJ.m-2.day-1] net solar or shortwave radiation
        albedo (float): [-] canopy reflection coefficient

    Returns:
        Rns (float):
    """
    return (1 - albedo) * rs


def eq39(t_min, t_max, ea, rs, rso):
    """Net outgoing longwave radiation

    Args:
        t_min (float): [K] min absolute temperature during the day
        t_max (float):  [K] max absolute temperature during the day
        ea (float): [kPa] actual vapor pressure
        rs (float): [MJ.m-2.day-1] solar radiation
        rso (float): [MJ.m-2.day-1] clear-sky radiation

    Returns:
        Rnl (float): [MJ.m-2.day-1]
    """
    alpha = 4.903e-9  # [MJ.K-4.m-2.day-1] Stefan-Boltzmann constant
    return alpha * (t_max ** 4 + t_min ** 4) / 2 * (0.34 - 0.14 * sqrt(ea)) * (1.35 * rs / rso - 0.35)



def eq53(rn, g, temp, ws, es, ea, delta, gamma):
    """Hourly reference evapotranspiration

    Args:
        rn (float): [MJ.m-2.h-1] net radiation at the crop surface
        g (float): [MJ.m-2.h-1] soil heat flux density
        temp (float): [°C] mean air temperature at 2 m height
        ws (float): [m.s-1] wind speed at 2 m height
        es (float): [kPa] saturation vapor pressure
        ea (float): [kPa] actual vapor pressure
        delta (float): [kPa.°C-1] slope relation vapor pressure temperature
        gamma (float): [kPa.°C-1] psychrometric constant

    Returns:
        eto (float): [mm.day-1] reference evapotranspiration
    """
    num = 0.408 * delta * (rn - g) + gamma * 37 / (temp + 273) * ws * (es - ea)
    denom = delta + gamma * (1 + 0.34 * ws)

    return num / denom


def eq54(es, rh):
    """Actual vapour pressure.

    Args:
        es (float): [kPa] saturation vapor pressure
        rh (float): [%] relative humidity

    Returns:
        ea (float): [kPa] actual vapor pressure
    """
    return es * rh / 100



from datetime import datetime
from math import radians

import pytest
from fao56.chap3 import eq11, eq13, eq21, eq23, eq24, eq25, eq34, eq38, eq39, eq8


def test_eqs_compare_to_ex17():
    date = datetime(1998, 4, 15)
    latitude = 13 + 44 / 60  # 50°48’N
    t_max = 34.8  # [°C]
    t_min = 25.6  # [°C]
    ea = 2.85  # [kPa]

    atm_pressure = 101.3  # [kPa]

    t_mean = (t_max + t_min) / 2
    delta = eq13(t_mean)
    assert delta == pytest.approx(0.246, abs=1e-3)

    gamma = eq8(atm_pressure)
    assert gamma == pytest.approx(0.0674, abs=1e-4)

    es_max = eq11(t_max)
    assert es_max == pytest.approx(5.56, abs=1e-2)

    es_min = eq11(t_min)
    assert es_min == pytest.approx(3.28, abs=1e-2)

    doy = (date - datetime(date.year, 1, 1)).days + 1
    dr = eq23(doy)
    decli = eq24(doy)
    sunset_hour = eq25(radians(latitude), decli)

    ra = eq21(dr, sunset_hour, radians(latitude), decli)
    assert ra == pytest.approx(38.06, abs=1e-2)

    n = eq34(sunset_hour)
    assert n == pytest.approx(12.31, abs=1e-1)

    rs = 22.65  # [MJ.m-2.day-1]
    rns = eq38(rs)
    assert rns == pytest.approx(17.44, abs=1e-2)

    rso = 28.54  # [MJ.m-2.day-1]
    rnl = eq39(t_min + 273.15, t_max + 273.15, ea, rs, rso)
    assert rnl == pytest.approx(3.11, abs=1e-2)


def test_eqs_compare_to_ex18():
    date = datetime(1998, 7, 6)
    latitude = 50 + 48 / 60  # 50°48’N
    t_max = 21.5  # [°C]
    t_min = 12.3  # [°C]

    atm_pressure = 100.1  # [kPa]

    t_mean = (t_max + t_min) / 2
    delta = eq13(t_mean)
    assert delta == pytest.approx(0.122, abs=1e-3)

    gamma = eq8(atm_pressure)
    assert gamma == pytest.approx(0.0666, abs=1e-4)

    es_max = eq11(t_max)
    assert es_max == pytest.approx(2.564, abs=1e-3)

    es_min = eq11(t_min)
    assert es_min == pytest.approx(1.431, abs=1e-3)

    doy = (date - datetime(date.year, 1, 1)).days + 1
    dr = eq23(doy)
    decli = eq24(doy)
    sunset_hour = eq25(radians(latitude), decli)

    ra = eq21(dr, sunset_hour, radians(latitude), decli)
    assert ra == pytest.approx(41.09, abs=1e-2)

    n = eq34(sunset_hour)
    assert n == pytest.approx(16.1, abs=1e-1)

    rs = 22.07  # [MJ.m-2.day-1]
    rns = eq38(rs)
    assert rns == pytest.approx(17.0, abs=1e-2)

    ea = 1.409  # [kPa]
    rso = 30.9  # [MJ.m-2.day-1]
    rnl = eq39(t_min + 273.15, t_max + 273.15, ea, rs, rso)
    assert rnl == pytest.approx(3.71, abs=1e-2)



"""
Cleaner API for some important equations in the book
"""
from datetime import datetime
from math import pi, radians, sqrt

from .chap2 import eq6
from .chap3 import eq11, eq13, eq21, eq23, eq24, eq25, eq28, eq29, eq30, eq32, eq37, eq38, eq39, eq7, eq8
from .chap4 import eq53, eq54


def sat_vap(temp):
    """Saturated vapor pressure

    Args:
        temp (float): [°C] air or organtemperature

    Returns:
        (float): [kPa] Saturated vapor pressure
    """
    return eq11(temp)


def vpd(t_air, rh, t_max=None, rh_max=None):
    """Vapour pressure deficit

    Args:
        t_air (float): [°C] average temperature (or min if t_max is given)
        rh (float): [%] average air humidity (or min is rh_max is given)
        t_max (float): [°C] maximum air temperature
        rh_max (float): [%] maximum air humidity

    Returns:
        vpd (float): [kPa]
    """
    t_min = t_air
    if t_max is None:
        t_max = t_air

    rh_min = rh
    if rh_max is None:
        rh_max = rh

    es_min = eq11(t_min)
    es_max = eq11(t_max)

    ea_min = eq54(es_min, rh_max)
    ea_max = eq54(es_max, rh_min)

    return (es_max + es_min) / 2 - (ea_max + ea_min) / 2


def rs_daily(date, latitude, altitude=0.):
    """Clear sky daily irradiance

    Args:
        date (datetime): Day of computation
        latitude (float): [°] Latitude
        altitude (float): [m] altitude of place

    Returns:
        rso (float): [W.m-2] average radiation during 24h
    """
    doy = (date - datetime(date.year, 1, 1)).days + 1
    dr = eq23(doy)
    decli = eq24(doy)
    sunset_hour = eq25(radians(latitude), decli)

    ra = eq21(dr, sunset_hour, radians(latitude), decli)
    return eq37(ra, altitude) * 1e6 / (24 * 3600)


def rs_hourly(date, latitude, longitude, altitude=0., sunset=None):
    """Clear sky daily irradiance

    Args:
        date (datetime): UTC, beginning of interval, assuming solar time if longitude=0
        latitude (float): [°] Latitude
        longitude (float): [°] Longitude
        altitude (float): [m] altitude of place
        sunset (float): [rad] sunset hour angle, if None, will be computed

    Returns:
        rso (float): [W.m-2] average radiation during 1h
    """
    day = date.replace(hour=0, minute=0, second=0, microsecond=0)
    doy = (day - datetime(date.year, 1, 1)).days + 1
    dr = eq23(doy)
    decli = eq24(doy)
    if sunset is None:
        sunset = eq25(radians(latitude), decli)

    sunrise = -sunset  # by definition in hour angles

    t = (date - day).total_seconds() / 3600 + 0.5  # mid interval
    sc = eq32(doy)
    omega = pi / 12 * ((t + 24 * longitude / 360 + sc) - 12)  # eq31
    omega1 = eq29(omega, 1)
    omega2 = eq30(omega, 1)
    if omega2 <= sunrise:
        return 0.

    if omega1 >= sunset:
        return 0.

    omega1 = max(sunrise, omega1)
    omega2 = min(sunset, omega2)

    ra = eq28(dr, omega1, omega2, radians(latitude), decli)

    return eq37(ra, altitude) * 1e6 / 3600


def eto_daily(date, latitude, altitude, t_air, rh, rs, ws, t_max=None, rh_max=None, atm_pressure=None):
    """Daily reference evapotranspiration.

    Args:
        date (datetime): Day of computation
        latitude (float): [°] Latitude
        altitude (float): [m] altitude of place
        t_air (float): [°C] average temperature (or min if t_max is given)
        rh (float): [%] average air humidity (or min is rh_max is given)
        rs (float): [W.m-2] Daily average solar radiation
        ws (float): [m.s-1] Wind speed at 2 meters above ground
        t_max (float): [°C] maximum air temperature
        rh_max (float): [%] maximum air humidity
        atm_pressure (float): [kPa] Atmospheric pressure

    Returns:
        ETo (float): [mm.day-1] Reference evapotranspiration
    """
    t_min = t_air
    if t_max is None:
        t_max = t_air

    rh_min = rh
    if rh_max is None:
        rh_max = rh

    if atm_pressure is None:
        atm_pressure = eq7(altitude)

    # psychrometric_constant
    gamma = eq8(atm_pressure)

    # Saturation vapor pressure
    es_min = eq11(t_min)
    es_max = eq11(t_max)
    es = (es_min + es_max) / 2.

    # Actual vapor pressure in the air
    ea = (eq54(es_min, rh_max) + eq54(es_max, rh_min)) / 2

    t_mean = (t_max + t_min) / 2

    # Derivative of es [kPa.°C-1]
    delta = eq13(t_mean)

    # net radiation
    rso = rs_daily(date, latitude, altitude) * 24 * 3600e-6  # [W.m-2] to [MJ.m-2.day-1]
    rs = min(rs * 24 * 3600e-6, rso)  # limit measured radiation to clear sky theoretical one
    rns = eq38(rs)

    rnl = eq39(t_min + 273.15, t_max + 273.15, ea, rs, rso)
    rn = max(0., rns - rnl)
    g = 0.

    return eq6(rn, g, t_mean, ws, es, ea, delta, gamma)


def eto_hourly(date, latitude, longitude, altitude, t_air, rh, rs, ws, g=0., atm_pressure=None):
    """Daily reference evapotranspiration.

    Args:
        date (datetime): UTC, beginning of interval, assuming solar time if longitude=0
        latitude (float): [°] Latitude
        longitude (float): [°] Longitude
        altitude (float): [m] altitude of place
        t_air (float): [°C] Mean air temperature
        rh (float): [%] Hourly mean relative humidity (between 0 and 100)
        rs (float): [W.m-2] Hourly average solar radiation
        ws (float): [m.s-1] Wind speed at 2 meters above ground
        g (float): [W.m-2] Ground heat flux
        atm_pressure (float): [kPa] Atmospheric pressure

    Returns:
        ETo (float): [mm.h-1] Reference evapotranspiration
    """
    if atm_pressure is None:
        atm_pressure = eq7(altitude)

    # psychrometric_constant
    gamma = eq8(atm_pressure)

    # Saturation vapor pressure
    es = eq11(t_air)

    # Actual vapor pressure in the air
    ea = eq54(es, rh)

    # Derivative of es [kPa.°C-1]
    delta = eq13(t_air)

    # net radiation
    rso = rs_hourly(date, latitude, longitude, altitude) * 3600e-6  # [W.m-2] to [MJ.m-2.h-1]
    rs = min(rs * 3600e-6, rso)  # limit measured radiation to clear sky theoretical one
    rns = eq38(rs)
    if rso < 1e-6:
        sky_cover_effect = 1.
    else:
        sky_cover_effect = (1.35 * rs / rso - 0.35)

    sigma = 2.043e-10  # [MJ.K-4.m-2.h-1] Stefan-Boltzmann constant
    # eq39 modified for hourly
    rnl = sigma * (t_air + 273.15) ** 4 * (0.34 - 0.14 * sqrt(ea)) * sky_cover_effect

    rn = max(0., rns - rnl)

    return eq53(rn, g * 3600e-6, t_air, ws, es, ea, delta, gamma)



from datetime import datetime, timedelta
from pathlib import Path

import pytest
from pandas import read_csv

from fao56.clean import eto_daily, eto_hourly, rs_daily, rs_hourly, vpd


def test_eto_daily_compare_to_fig18():
    # longitude = 12.11
    latitude = -5.33
    altitude = 20  # [m]

    # read table
    table_pth = Path(__file__).parent / "etp_table_chap4_fig18.txt"
    with open(table_pth) as f:
        lines = [line.strip() for line in f.readlines() if len(line.strip()) > 0]

    # perform comparison
    for month_ind, line in enumerate(lines[2:]):
        month, t_min, t_max, humidity, wind, sunshine, radiation, eto_ref = line.split()
        date = datetime(2016, month_ind + 1, 1)
        eto = eto_daily(date,
                        latitude,
                        altitude,
                        float(t_min),
                        float(humidity),
                        float(radiation) * 1e6 / (24 * 3600),  # [MJ.m-2.day-1] to [W.m-2]
                        float(wind) * 1000 / (24 * 3600),  # [km.day-1] to [m.s-1]
                        float(t_max))

        assert eto == pytest.approx(float(eto_ref), abs=0.18)


def test_vpd_compare_to_ex18():
    # example 18, FAO, chap.4
    t_max = 21.5
    t_min = 12.3
    rh_max = 84
    rh_min = 63

    assert vpd(t_min, rh_min, t_max, rh_max) == pytest.approx(0.589, abs=1e-3)


def test_vpd_compare_to_ex19():
    # example 19, FAO, chap.4
    t_air = 28
    rh = 90
    assert vpd(t_air, rh) == pytest.approx(0.378, abs=1e-3)

    t_air = 38
    rh = 52
    assert vpd(t_air, rh) == pytest.approx(3.180, abs=1e-3)


def test_rs_daily_compare_to_ex18():
    # example 18, FAO, chap.4
    day = datetime(2017, 7, 6)
    latitude = 50.8
    altitude = 100
    rso = rs_daily(day, latitude, altitude)

    rs_ref = 30.90 * 1e6 / (24 * 3600)  # [MJ.m-2.day-1] to [W.m-2]
    assert rso == pytest.approx(rs_ref, abs=1e-1)


def test_rs_hourly_compare_to_ex19():
    # example 19, FAO, chap.4
    latitude = 16 + 13 / 60  # 16°13'N
    longitude = -(16 + 15 / 60)  # 16°15'W
    altitude = 8

    date = datetime(2017, 10, 1, 14) + timedelta(hours=1)  # correction for time zone wrong in fao
    rso = rs_hourly(date, latitude, longitude, altitude)
    rs_ref = 2.658 * 1e6 / 3600  # [MJ.m-2.h-1] to [W.m-2]
    assert rso == pytest.approx(rs_ref, abs=1e-1)

    date = datetime(2017, 10, 1, 2) + timedelta(hours=1)  # correction for time zone wrong in fao
    rso = rs_hourly(date, latitude, longitude, altitude)
    assert rso == pytest.approx(0, abs=1e-1)


def test_eto_daily_compare_to_ex18():
    # example 18, FAO, chap.4
    t_max = 21.5
    t_min = 12.3
    rh_max = 84
    rh_min = 63

    ws = 2.078
    day = datetime(2017, 7, 6)
    latitude = 50.8
    altitude = 100  # [m]
    rs = 22.07 * 1e6 / (24 * 3600)  # [MJ.m-2.day-1] to [W.m-2]

    eto = eto_daily(day, latitude, altitude, t_min, rh_min, rs, ws, t_max, rh_max)
    assert eto == pytest.approx(3.88, abs=1e-2)


def test_eto_daily_handles_bad_radiation_values():
    t_max = 21.5
    t_min = 12.3
    rh = 75
    ws = 2.078
    day = datetime(2017, 7, 6)
    latitude = 50.8
    altitude = 100  # [m]
    rso = rs_daily(day, latitude, altitude)

    eto_ref = eto_daily(day, latitude, altitude, t_min, rh, rso, ws, t_max)

    # check that etp is cropped for too big values of rg
    for s in (0.1, 0.5, 0.99, 1., 1.1, 2.):
        eto = eto_daily(day, latitude, altitude, t_min, rh, rso * s, ws, t_max)
        assert eto <= eto_ref


def test_eto_daily_is_always_positive():
    latitude = 45
    df_climate = read_csv(Path(__file__).parent / "climate.csv", sep=";", comment="#", index_col='date',
                          parse_dates=True)

    for date, w in df_climate.iterrows():
        rs = w["rg"] * 1e4 / 3600  # [J.cm-2] to [W.m-2]
        assert eto_daily(date, latitude, 0., w["tmin"], w["rh"], rs, w["ws"], w["tmax"]) >= 0


def test_eto_hourly_compare_to_ex19():
    # example 19, FAO, chap.4
    latitude = 16 + 13 / 60  # 16°13'N
    longitude = -(16 + 15 / 60)  # 16°15'W
    altitude = 8

    date = datetime(2017, 10, 1, 14) + timedelta(hours=1)  # correction for time zone wrong in fao

    t_air = 38  # [°C]
    rh = 52  # [%}
    ws = 3.3  # [m.s-1]
    rs = 2.450 * 1e6 / 3600  # [MJ.m-2.h-1] to [W.m-2]
    g = 0.175 * 1e6 / 3600  # [MJ.m-2.h-1] to [W.m-2]

    eto = eto_hourly(date, latitude, longitude, altitude, t_air, rh, rs, ws, g=g)
    assert eto == pytest.approx(0.63, abs=1e-2)



"""
Retrieve weather for the period from 1950 till 2100
Select days with maximum and minimum features
"""
from datetime import datetime
from pathlib import Path

from drias.climate import fetch_climate

# get weather 150 years
df = fetch_climate(latitude=43.520199,
                   longitude=5.647930,
                   scenario="med",
                   date_beg=datetime(1950, 1, 1),
                   date_end=datetime(2099, 12, 31))

df = df.rename(columns={
    'maxTemperature': 'tmax',
    'minTemperature': 'tmin',
    'meanTemperature': 'tmean',
    'relativeHumidity': 'rh',
    'solarRadiation': 'rg',
    'totalPrecipitation': 'rain',
    'windSpeed': 'ws'})

# extract days with minimum or maximum values
extrema = set(df.idxmin()) | set(df.idxmax())

df_extrema = df.loc[extrema].sort_index().sort_index(axis=1)

# write resulting dataframe
with open(Path(__file__).parent / "climate.csv", 'w', encoding="utf8") as fhw:
    fhw.write(f"# This file has been generated by {__file__}\n#\n")
    fhw.write("# rain: [mm] total precipitation for the day\n")
    fhw.write("# rg: [J cm-2] solar radiation\n")
    fhw.write("# rh: [-] relative humidity between 0 and 1\n")
    fhw.write("# tmax: [°C] maximum temperature for a day\n")
    fhw.write("# tmean: [°C] mean temperature for a day\n")
    fhw.write("# tmin: [°C] minimum temperature for a day\n")
    fhw.write("# ws: [m s-1] mean wind speed\n")
    fhw.write("#\n")
    df_extrema.to_csv(fhw, sep=";")
